import os
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, cast

import google.cloud.bigquery as bigquery  # noqa: I250
from google.api_core import exceptions as api_core_exceptions
from google.api_core import retry
from google.cloud import exceptions as gcp_exceptions
from google.cloud.bigquery.retry import _RETRYABLE_REASONS

from dlt.common import logger
from dlt.common.destination import DestinationCapabilitiesContext, PreparedTableSchema
from dlt.common.destination.client import (
    HasFollowupJobs,
    FollowupJobRequest,
    RunnableLoadJob,
    SupportsStagingDestination,
    LoadJob,
)
from dlt.common.json import json
from dlt.common.runtime.signals import sleep
from dlt.common.schema import TColumnSchema, Schema, TTableSchemaColumns
from dlt.common.schema.typing import TColumnType
from dlt.common.schema.utils import get_inherited_table_hint, get_columns_names_with_prop
from dlt.common.storages.load_package import destination_state
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.common.typing import DictStrAny
from dlt.common.data_writers.escape import escape_bigquery_literal
from dlt.destinations.exceptions import (
    DatabaseTransientException,
    DatabaseUndefinedRelation,
    DestinationSchemaWillNotUpdate,
    DestinationTerminalException,
    DatabaseTerminalException,
    LoadJobTerminalException,
)
from dlt.destinations.impl.bigquery.bigquery_adapter import (
    AUTODETECT_SCHEMA_HINT,
    CLUSTER_COLUMNS_HINT,
    PARTITION_HINT,
    PARTITION_EXPIRATION_DAYS_HINT,
    CLUSTER_HINT,
    TABLE_DESCRIPTION_HINT,
    ROUND_HALF_EVEN_HINT,
    ROUND_HALF_AWAY_FROM_ZERO_HINT,
    TABLE_EXPIRATION_HINT,
    should_autodetect_schema,
)
from dlt.destinations.impl.bigquery.configuration import BigQueryClientConfiguration
from dlt.destinations.impl.bigquery.warnings import per_column_cluster_hint_deprecated
from dlt.destinations.impl.bigquery.sql_client import BigQuerySqlClient, BQ_TERMINAL_REASONS
from dlt.destinations.job_client_impl import SqlJobClientWithStagingDataset
from dlt.destinations.job_impl import DestinationJsonlLoadJob, DestinationParquetLoadJob
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.sql_client import SqlClientBase


class BigQueryLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(
        self,
        file_path: str,
        http_timeout: float,
        retry_deadline: float,
    ) -> None:
        super().__init__(file_path)
        self._default_retry = bigquery.DEFAULT_RETRY.with_deadline(retry_deadline)
        self._http_timeout = http_timeout
        self._job_client: "BigQueryClient" = None
        self._bq_load_job: bigquery.LoadJob = None
        # vars only used for testing
        self._created_job = False
        self._resumed_job = False

    def run(self) -> None:
        # start the job (or retrieve in case it already exists)
        try:
            self._bq_load_job = self._job_client._create_load_job(self._load_table, self._file_path)
            self._created_job = True
        except api_core_exceptions.GoogleAPICallError as gace:
            reason = BigQuerySqlClient._get_reason_from_errors(gace)
            if reason == "notFound":
                # google.api_core.exceptions.NotFound: 404 – table not found
                raise DatabaseUndefinedRelation(gace) from gace
            elif (
                reason == "duplicate"
            ):  # google.api_core.exceptions.Conflict: 409 PUT – already exists
                self._bq_load_job = self._job_client._retrieve_load_job(self._file_path)
                self._resumed_job = True
                logger.info(
                    f"Found existing bigquery job for job {self._file_name}, will resume job."
                )
            elif reason in BQ_TERMINAL_REASONS:
                # google.api_core.exceptions.BadRequest - will not be processed ie bad job name
                raise LoadJobTerminalException(
                    self._file_path, f"The server reason was: `{reason}`"
                ) from gace
            else:
                raise DatabaseTransientException(gace) from gace

        # we loop on the job thread until we detect a status change
        while True:
            sleep(1)
            # not done yet
            if not self._bq_load_job.done(retry=self._default_retry, timeout=self._http_timeout):
                continue
            # done, break loop and go to completed state
            if self._bq_load_job.output_rows is not None and self._bq_load_job.error_result is None:
                break
            reason = self._bq_load_job.error_result.get("reason")
            if reason in BQ_TERMINAL_REASONS:
                # the job permanently failed for the reason above
                raise DatabaseTerminalException(
                    Exception(
                        f"Bigquery Load Job failed, reason reported from bigquery: `{reason}`"
                    )
                )
            elif reason in ["internalError"]:
                logger.warning(
                    f"Got reason `{reason}` for job `{self._file_name}`, job considered still"
                    f" running. ({self._bq_load_job.error_result})"
                )
                continue
            else:
                raise DatabaseTransientException(
                    Exception(
                        "Bigquery Job needs to be retried, reason reported from bigquery"
                        f" `{reason}`"
                    )
                )

    def failed_message(self) -> str:
        if self._bq_load_job:
            return json.dumps(
                {
                    "error_result": self._bq_load_job.error_result,
                    "errors": self._bq_load_job.errors,
                    "job_start": self._bq_load_job.started,
                    "job_end": self._bq_load_job.ended,
                    "job_id": self._bq_load_job.job_id,
                }
            )
        return super().failed_message()

    def exception(self) -> BaseException:
        if self._bq_load_job:
            return self._bq_load_job.exception()  # type: ignore[no-any-return]
        return super().exception()

    @staticmethod
    def get_job_id_from_file_path(file_path: str) -> str:
        return Path(file_path).name.replace(".", "_")


class BigQueryMergeJob(SqlMergeFollowupJob):
    @classmethod
    def _gen_table_setup_clauses(
        cls, table_chain: Sequence[PreparedTableSchema], sql_client: SqlClientBase[Any]
    ) -> List[str]:
        """generate final tables from staging table schema for autodetect tables"""
        sql: List[str] = []
        for table in table_chain:
            if should_autodetect_schema(table):
                table_name, staging_table_name = sql_client.get_qualified_table_names(table["name"])
                sql.append(f"CREATE TABLE IF NOT EXISTS {table_name} LIKE {staging_table_name}")
        return sql

    @classmethod
    def gen_key_table_clauses(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        key_clauses: Sequence[str],
        for_delete: bool,
    ) -> List[str]:
        sql: List[str] = [
            f"FROM {root_table_name} AS d WHERE EXISTS (SELECT 1 FROM {staging_root_table_name} AS"
            f" s WHERE {clause.format(d='d', s='s')})"
            for clause in key_clauses
        ]
        return sql


class BigQueryClient(SqlJobClientWithStagingDataset, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: BigQueryClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = SqlJobClientWithStagingDataset.create_dataset_names(
            schema, config
        )
        sql_client = BigQuerySqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
            config.get_location(),
            config.project_id,
            config.http_timeout,
            config.retry_deadline,
        )
        super().__init__(schema, config, sql_client)
        self.config: BigQueryClientConfiguration = config
        self.sql_client: BigQuerySqlClient = sql_client  # type: ignore
        self.type_mapper = self.capabilities.get_type_mapper()

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [BigQueryMergeJob.from_table_chain(table_chain, self.sql_client)]

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        truncate_tables = truncate_tables or []

        # split array into tables that have autodetect schema and those that don't
        autodetect_tables = [
            t for t in truncate_tables if should_autodetect_schema(self.prepare_load_table(t))
        ]
        non_autodetect_tables = [t for t in truncate_tables if t not in autodetect_tables]

        # if any table has schema autodetect, we need to make sure to only truncate tables that exist
        super().initialize_storage(truncate_tables=non_autodetect_tables)
        self.sql_client.truncate_tables_if_exist(*autodetect_tables)

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id)

        if not job:
            insert_api = table.get("x-insert-api", "default")
            if insert_api == "streaming":
                if table["write_disposition"] != "append":
                    raise DestinationTerminalException(
                        "BigQuery streaming insert can only be used with"
                        " `write_disposition='append'`. Resource received"
                        f" `write_disposition={table['write_disposition']}`"
                    )
                parsed_file = ParsedLoadJobFileName.parse(file_path)
                if parsed_file.file_format in ["jsonl", "typed-jsonl"]:
                    job_cls = DestinationJsonlLoadJob
                elif parsed_file.file_format == "parquet":
                    job_cls = DestinationParquetLoadJob  # type: ignore
                else:
                    raise ValueError(
                        f"Unsupported file type for BigQuery streaming inserts: `{file_path}`"
                    )

                job = job_cls(
                    file_path,
                    self.config,  # type: ignore
                    destination_state(),
                    _streaming_load,  # type: ignore
                    [],
                    callable_requires_job_client_args=True,
                )
            else:
                job = BigQueryLoadJob(
                    file_path,
                    self.config.http_timeout,
                    self.config.retry_deadline,
                )
        return job

    def _bigquery_partition_clause(self, partition_hint: Optional[Dict[str, str]]) -> str:
        """Generate partition clause for BigQuery SQL.

        Args:
            partition_hint (Optional[Dict[str, str]]): Partition hint.

        Returns:
            str: The partition clause for BigQuery SQL.
        """
        if not partition_hint:
            return ""

        [(column_name, clause)] = partition_hint.items()
        return (
            "PARTITION BY"
            f" {clause.format(column_name=self.sql_client.escape_column_name(column_name))}"
        )

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        # Return empty columns which will skip table CREATE or ALTER to let BigQuery
        # auto-detect table from data.
        table = self.prepare_load_table(table_name)
        if should_autodetect_schema(table):
            return []

        sql = super()._get_table_update_sql(table_name, new_columns, generate_alter)
        canonical_name = self.sql_client.make_qualified_table_name(table_name)

        # partition and cluster clauses are only valid for CREATE TABLE, not ALTER TABLE
        if not generate_alter:
            # handle partitioning when user passes a string to the `partition` param in bigquery_adapter
            if partition_list := [
                c for c in new_columns if c.get("partition") or c.get(PARTITION_HINT, False)
            ]:
                if len(partition_list) > 1:
                    col_names = [
                        self.sql_client.escape_column_name(c["name"]) for c in partition_list
                    ]
                    raise DestinationSchemaWillNotUpdate(
                        canonical_name, col_names, "Partition requested for more than one column"
                    )
                elif (c := partition_list[0])["data_type"] == "date":
                    sql[0] += f"\nPARTITION BY {self.sql_client.escape_column_name(c['name'])}"
                elif (c := partition_list[0])["data_type"] == "timestamp":
                    sql[0] = (
                        f"{sql[0]}\nPARTITION BY"
                        f" DATE({self.sql_client.escape_column_name(c['name'])})"
                    )
                # Automatic partitioning of an INT64 type requires us to be prescriptive - we treat the column as a UNIX timestamp.
                # This is due to the bounds requirement of GENERATE_ARRAY function for partitioning.
                # The 10,000 partitions limit makes it infeasible to cover the entire `bigint` range.
                # The array bounds, with daily partitions (86400 seconds in a day), are somewhat arbitrarily chosen.
                # See: https://dlthub.com/devel/dlt-ecosystem/destinations/bigquery#supported-column-hints
                elif (c := partition_list[0])["data_type"] == "bigint":
                    sql[0] += (
                        "\nPARTITION BY"
                        f" RANGE_BUCKET({self.sql_client.escape_column_name(c['name'])},"
                        " GENERATE_ARRAY(-172800000, 691200000, 86400))"
                    )
            # handle partitioning when user passes a PartitionTransformation to the `partition` param in bigquery_adapter
            partition_hint = table.get(PARTITION_HINT)
            if isinstance(partition_hint, dict) and len(partition_hint) > 1:
                col_names = [
                    self.sql_client.escape_column_name(col) for col, v in partition_hint.items()
                ]
                raise DestinationSchemaWillNotUpdate(
                    canonical_name, col_names, "Partition requested for more than one column"
                )
            sql[0] += self._bigquery_partition_clause(
                partition_hint if isinstance(partition_hint, dict) else None
            )

        # Collect cluster columns from table-level and per-column hints
        cluster_columns_from_table_hint = list(
            cast(Iterable[str], table.get(CLUSTER_COLUMNS_HINT, []))
        )
        cluster_columns_from_column_hints = [
            c["name"] for c in new_columns if c.get("cluster") or c.get(CLUSTER_HINT, False)
        ]

        # Deprecation warning for per-column cluster hints
        if cluster_columns_from_column_hints and not cluster_columns_from_table_hint:
            per_column_cluster_hint_deprecated(cluster_columns_from_column_hints)

        # Prefer table-level cluster columns if present, otherwise fallback to per-column hints
        cluster_columns_final = (
            cluster_columns_from_table_hint
            if cluster_columns_from_table_hint
            else cluster_columns_from_column_hints
        )

        if cluster_columns_final and not generate_alter:
            cluster_list = [
                self.sql_client.escape_column_name(col) for col in cluster_columns_final
            ]
            sql[0] += "\nCLUSTER BY " + ", ".join(cluster_list)

        # Table options.
        table_options: DictStrAny = {
            "description": (
                f"'{table.get(TABLE_DESCRIPTION_HINT)}'"
                if table.get(TABLE_DESCRIPTION_HINT)
                else None
            ),
            "expiration_timestamp": (
                f"TIMESTAMP '{table.get(TABLE_EXPIRATION_HINT)}'"
                if table.get(TABLE_EXPIRATION_HINT)
                else None
            ),
            "partition_expiration_days": (
                str(table.get(PARTITION_EXPIRATION_DAYS_HINT))
                if table.get(PARTITION_EXPIRATION_DAYS_HINT)
                else None
            ),
        }
        if not any(table_options.values()):
            return sql

        if generate_alter:
            logger.info(
                f"Table options for {table_name} are not applied on ALTER TABLE. Make sure that you"
                " set the table options ie. by using bigquery_adapter, before it is created."
            )
        else:
            sql[0] += (
                "\nOPTIONS ("
                + ", ".join(
                    [f"{key}={value}" for key, value in table_options.items() if value is not None]
                )
                + ")"
            )

        return sql

    def prepare_load_table(self, table_name: str) -> Optional[PreparedTableSchema]:
        table = super().prepare_load_table(table_name)
        if table_name not in self.schema.dlt_table_names():
            if TABLE_DESCRIPTION_HINT not in table:
                table[TABLE_DESCRIPTION_HINT] = (  # type: ignore[name-defined, typeddict-unknown-key, unused-ignore]
                    get_inherited_table_hint(
                        self.schema.tables, table_name, TABLE_DESCRIPTION_HINT, allow_none=True
                    )
                )
            if AUTODETECT_SCHEMA_HINT not in table:
                table[AUTODETECT_SCHEMA_HINT] = (  # type: ignore[typeddict-unknown-key]
                    get_inherited_table_hint(
                        self.schema.tables, table_name, AUTODETECT_SCHEMA_HINT, allow_none=True
                    )
                    or self.config.autodetect_schema
                )

        return table

    def get_storage_tables(
        self, table_names: Iterable[str]
    ) -> Iterable[Tuple[str, TTableSchemaColumns]]:
        """Gets table schemas from BigQuery using INFORMATION_SCHEMA or get_table for hidden datasets"""
        if not self.sql_client.is_hidden_dataset:
            return super().get_storage_tables(table_names)

        # use the api to get storage tables for hidden dataset
        schema_tables: List[Tuple[str, TTableSchemaColumns]] = []
        for table_name in table_names:
            try:
                schema_table: TTableSchemaColumns = {}
                table = self.sql_client.native_connection.get_table(
                    self.sql_client.make_qualified_table_name(table_name, quote=False),
                    retry=self.sql_client._default_retry,
                    timeout=self.config.http_timeout,
                )
                for c in table.schema:
                    schema_c: TColumnSchema = {
                        "name": c.name,
                        "nullable": c.is_nullable,
                        **self._from_db_type(c.field_type, c.precision, c.scale),
                    }
                    schema_table[c.name] = schema_c
                schema_tables.append((table_name, schema_table))
            except gcp_exceptions.NotFound:
                # table is not present
                schema_tables.append((table_name, {}))
        return schema_tables

    def _get_info_schema_columns_query(
        self, catalog_name: Optional[str], schema_name: str, folded_table_names: List[str]
    ) -> Tuple[str, List[Any]]:
        """Bigquery needs to scope the INFORMATION_SCHEMA.COLUMNS with project and dataset name so standard query generator cannot be used."""
        # escape schema and catalog names
        catalog_name = self.capabilities.escape_identifier(catalog_name)
        schema_name = self.capabilities.escape_identifier(schema_name)

        query = f"""
SELECT {",".join(self._get_storage_table_query_columns())}
    FROM {catalog_name}.{schema_name}.INFORMATION_SCHEMA.COLUMNS
"""
        if folded_table_names:
            # placeholder for each table
            table_placeholders = ",".join(["%s"] * len(folded_table_names))
            query += f"WHERE table_name IN ({table_placeholders}) "
        query += "ORDER BY table_name, ordinal_position"

        return query, folded_table_names

    def _get_column_def_sql(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        column_def_sql = super()._get_column_def_sql(column, table)

        # generate additional column options clause
        # see: https://docs.cloud.google.com/bigquery/docs/reference/standard-sql/data-definition-language#alter_column_set_options_statement
        options = []
        if column.get(ROUND_HALF_EVEN_HINT, False):
            options.append("rounding_mode='ROUND_HALF_EVEN'")
        if column.get(ROUND_HALF_AWAY_FROM_ZERO_HINT, False):
            options.append("rounding_mode='ROUND_HALF_AWAY_FROM_ZERO'")
        if column.get("description", False):
            escaped_description = escape_bigquery_literal(column.get("description"))
            options.append(f"description={escaped_description}")

        if options:
            option_arguments = ", ".join(options)
            option_str = f" OPTIONS ({option_arguments})"
            column_def_sql += option_str
        return column_def_sql

    def _create_load_job(self, table: PreparedTableSchema, file_path: str) -> bigquery.LoadJob:
        # append to table for merge loads (append to stage) and regular appends.
        table_name = table["name"]

        # determine whether we load from local or url
        bucket_path = None
        ext: str = os.path.splitext(file_path)[1][1:]
        if ReferenceFollowupJobRequest.is_reference_job(file_path):
            bucket_path = ReferenceFollowupJobRequest.resolve_reference(file_path)
            ext = os.path.splitext(bucket_path)[1][1:]

        # Select a correct source format
        source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
        decimal_target_types: Optional[List[str]] = None
        if ext == "parquet":
            source_format = bigquery.SourceFormat.PARQUET
            # parquet needs NUMERIC type auto-detection
            decimal_target_types = ["NUMERIC", "BIGNUMERIC"]

        job_id = BigQueryLoadJob.get_job_id_from_file_path(file_path)
        job_config = bigquery.LoadJobConfig(
            autodetect=False,
            write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
            create_disposition=bigquery.CreateDisposition.CREATE_NEVER,
            source_format=source_format,
            decimal_target_types=decimal_target_types,
            ignore_unknown_values=self.config.ignore_unknown_values,
            max_bad_records=0,
        )
        if should_autodetect_schema(table):
            # Allow BigQuery to infer and evolve the schema, note that dlt is not creating such tables at all.
            job_config = self._set_user_hints_with_schema_autodetection(table, job_config)

        if bucket_path:
            return self.sql_client.native_connection.load_table_from_uri(
                bucket_path,
                self.sql_client.make_qualified_table_name(table_name, quote=False),
                job_id=job_id,
                job_config=job_config,
                timeout=self.config.file_upload_timeout,
            )

        with open(file_path, "rb") as f:
            return self.sql_client.native_connection.load_table_from_file(
                f,
                self.sql_client.make_qualified_table_name(table_name, quote=False),
                job_id=job_id,
                job_config=job_config,
                timeout=self.config.file_upload_timeout,
            )

    def _set_user_hints_with_schema_autodetection(
        self, table: PreparedTableSchema, job_config: bigquery.LoadJobConfig
    ) -> bigquery.LoadJobConfig:
        job_config.autodetect = True
        job_config.ignore_unknown_values = self.config.ignore_unknown_values
        job_config.schema_update_options = bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION
        job_config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
        if partition_column_ := get_columns_names_with_prop(table, PARTITION_HINT):
            partition_column = partition_column_[0]
            col_dtype = table["columns"][partition_column]["data_type"]
            if col_dtype == "date":
                job_config.time_partitioning = bigquery.TimePartitioning(field=partition_column)
            elif col_dtype == "timestamp":
                job_config.time_partitioning = bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY, field=partition_column
                )
            elif col_dtype == "bigint":
                job_config.range_partitioning = bigquery.RangePartitioning(
                    field=partition_column,
                    range_=bigquery.PartitionRange(start=-172800000, end=691200000, interval=86400),
                )
        if clustering_columns := get_columns_names_with_prop(table, CLUSTER_HINT):
            job_config.clustering_fields = clustering_columns
        if table_description := table.get(TABLE_DESCRIPTION_HINT, False):
            job_config.destination_table_description = table_description
        if table_expiration := table.get(TABLE_EXPIRATION_HINT, False):
            raise ValueError(
                f"Table expiration time ({table_expiration}) can't be set with BigQuery type"
                " auto-detection enabled!"
            )
        return job_config

    def _retrieve_load_job(self, file_path: str) -> bigquery.LoadJob:
        job_id = BigQueryLoadJob.get_job_id_from_file_path(file_path)
        return cast(bigquery.LoadJob, self.sql_client.native_connection.get_job(job_id))

    def _from_db_type(
        self, bq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(bq_t, precision, scale)

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load


def _streaming_load(
    items: List[Dict[Any, Any]], table: Dict[str, Any], job_client: BigQueryClient
) -> None:
    """
    Upload the given items into BigQuery table, using streaming API.
    Streaming API is used for small amounts of data, with optimal
    batch size equal to 500 rows.

    Args:
        sql_client (dlt.destinations.impl.bigquery.bigquery.BigQueryClient):
            BigQuery client.
        items (List[Dict[Any, Any]]): List of rows to upload.
        table (Dict[Any, Any]): Table schema.
    """

    def _should_retry(exc: api_core_exceptions.GoogleAPICallError) -> bool:
        """Predicate to decide if we need to retry the exception.

        Args:
            exc (google.api_core.exceptions.GoogleAPICallError):
                Exception raised by the client.

        Returns:
            bool: True if the exception is retryable, False otherwise.
        """
        reason = exc.errors[0]["reason"]
        return reason in _RETRYABLE_REASONS

    sql_client = job_client.sql_client

    full_name = sql_client.make_qualified_table_name(table["name"], quote=False)

    bq_client = sql_client._client
    bq_client.insert_rows_json(
        full_name,
        items,
        retry=retry.Retry(predicate=_should_retry, deadline=600),  # with 10 mins deadline
    )



import re
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Set, Tuple
import yaml

from dlt.common.json import custom_encode, json

_HEADING_RE = re.compile(r"^(#{1,6})\s+(.*)")


class MarkdownHeading(NamedTuple):
    line_index: int  # line index in document
    depth: int  # number of # (1-6)
    text: str  # heading text after # chars


class MarkdownDocument:
    """Simple line-based markdown document for heading lookup, frontmatter, and line insertion."""

    def __init__(self, text: str) -> None:
        self._lines: List[str] = text.split("\n")

    def __str__(self) -> str:
        return "\n".join(self._lines)

    @property
    def lines(self) -> List[str]:
        return self._lines

    @property
    def is_empty(self) -> bool:
        return self._lines == [""]

    def find_headings(self, depth: Optional[int] = None) -> List[MarkdownHeading]:
        """Return all headings, optionally filtered by depth (1-6)."""
        result: List[MarkdownHeading] = []
        for i, line in enumerate(self._lines):
            m = _HEADING_RE.match(line.strip())
            if m:
                d = len(m.group(1))
                if depth is None or d == depth:
                    result.append(MarkdownHeading(i, d, m.group(2).strip()))
        return result

    def find_first_heading(self) -> Optional[MarkdownHeading]:
        """Return the first heading or None."""
        for i, line in enumerate(self._lines):
            m = _HEADING_RE.match(line.strip())
            if m:
                return MarkdownHeading(i, len(m.group(1)), m.group(2).strip())
        return None

    def find_line(self, text: str) -> Optional[int]:
        """Return index of first line whose stripped content equals `text`, or None."""
        for i, line in enumerate(self._lines):
            if line.strip() == text:
                return i
        return None

    def search(self, pattern: str) -> List[Tuple[int, "re.Match[str]"]]:
        """Return `(index, match)` for every line matching `pattern`."""
        compiled = re.compile(pattern)
        return [(i, m) for i, line in enumerate(self._lines) if (m := compiled.search(line))]

    def insert_lines(self, index: int, new_lines: List[str]) -> None:
        """Insert `new_lines` before `index`, mutating the document."""
        for j, entry in enumerate(new_lines):
            self._lines.insert(index + j, entry)

    def insert_md(self, index: Optional[int], other: "MarkdownDocument") -> None:
        """Insert all lines of `other` before `index`. `None` appends at end."""
        if index is None:
            index = len(self._lines)
        self.insert_lines(index, other.lines)

    def frontmatter(self) -> Tuple[Dict[str, Any], int]:
        """Parse `---` delimited YAML frontmatter.

        Returns `(data_dict, body_start_index)`. When there is no valid
        frontmatter, returns `({}, 0)`.
        """
        if not self._lines or self._lines[0].rstrip("\r") != "---":
            return {}, 0
        close = None
        for i in range(1, len(self._lines)):
            if self._lines[i].rstrip("\r") == "---":
                close = i
                break
        if close is None:
            return {}, 0
        fm_str = "\n".join(self._lines[1:close])
        data: Dict[str, Any] = yaml.safe_load(fm_str) or {}
        if not isinstance(data, dict):
            return {}, 0
        return data, close + 1

    @property
    def body_text(self) -> str:
        """Text after frontmatter (or full text when there is none)."""
        _, start = self.frontmatter()
        return "\n".join(self._lines[start:])

    @staticmethod
    def from_frontmatter(data: Dict[str, Any], body: str) -> "MarkdownDocument":
        """Create a document from a frontmatter dict and body text."""
        if not data:
            return MarkdownDocument(body)
        fm = yaml.dump(data, default_flow_style=False, sort_keys=False).rstrip("\n")
        return MarkdownDocument("---\n" + fm + "\n---\n" + body)


def parse_frontmatter(text: str) -> Tuple[Dict[str, Any], str]:
    """Split YAML frontmatter (`---` delimited) from markdown body."""
    doc = MarkdownDocument(text)
    data, body_start = doc.frontmatter()
    if not data and body_start == 0 and text:
        # no valid frontmatter — return original text as body
        return {}, text
    return data, doc.body_text


def render_frontmatter(data: Dict[str, Any], body: str) -> str:
    """Combine frontmatter dict and body into `---\\nyaml\\n---\\nbody`."""
    return str(MarkdownDocument.from_frontmatter(data, body))


def extract_first_heading(body: str) -> Optional[str]:
    """Return text of the first markdown heading, or None."""
    h = MarkdownDocument(body).find_first_heading()
    return h.text if h else None


def read_md_name_desc(path: Path) -> Tuple[str, str]:
    """Read a markdown file and return (name, description) from its frontmatter."""
    text = path.read_text(encoding="utf-8")
    try:
        fm, body = parse_frontmatter(text)
    except yaml.YAMLError:
        fm, body = {}, text
    name = fm.get("name", path.stem)
    desc = fm.get("description") or extract_first_heading(body) or ""
    return name, desc


# c0 controls (except tab 0x09, LF 0x0A, CR 0x0D), DEL, c1 controls
_CONTROL_CHAR_RE = re.compile("[\x00-\x08\x0b\x0c\x0e-\x1f\x7f\x80-\x9f]")


def sanitize_string(s: str) -> str:
    """Strip control characters and lone surrogates."""
    s = _CONTROL_CHAR_RE.sub("", s)
    s = s.encode("utf-8", errors="surrogatepass").decode("utf-8", errors="ignore")
    return s


def sanitize_value(val: Any) -> str:
    """Coerce a value to a sanitized string representation."""
    if val is None:
        return ""
    if isinstance(val, str):
        return sanitize_string(val)
    try:
        encoded = custom_encode(val)
    except TypeError:
        encoded = val
    return sanitize_string(str(encoded))


def _sanitize_json_value(val: Any) -> Any:
    """Sanitize a value for JSON output, preserving native JSON types."""
    if val is None:
        return None
    if isinstance(val, bool):
        return val
    if isinstance(val, (int, float)):
        return val
    if isinstance(val, str):
        return sanitize_string(val)
    try:
        encoded = custom_encode(val)
    except TypeError:
        encoded = val
    if isinstance(encoded, (dict, list)):
        return encoded
    return sanitize_string(str(encoded))


def render_md_table(
    columns: Sequence[str],
    rows: Sequence[Tuple[Any, ...]],
) -> str:
    """Render columns and rows as a Markdown table with sanitized values."""
    if not columns:
        return "*(empty result)*"

    headers = [sanitize_string(c) for c in columns]
    str_rows: List[List[str]] = [[sanitize_value(v) for v in row] for row in rows]

    lines = ["| " + " | ".join(headers) + " |"]
    lines.append("| " + " | ".join("---" for _ in headers) + " |")
    for str_row in str_rows:
        lines.append("| " + " | ".join(str_row) + " |")

    lines.append(f"\n*({len(str_rows)} row(s))*")
    return "\n".join(lines)


def render_jsonl(
    columns: Sequence[str],
    rows: Sequence[Tuple[Any, ...]],
) -> str:
    """Render columns and rows as JSON-lines (one JSON object per row)."""
    if not columns:
        return ""

    lines: List[str] = []
    col_list = list(columns)
    for row in rows:
        obj: Dict[str, Any] = {}
        for i, val in enumerate(row):
            if i < len(col_list):
                obj[col_list[i]] = _sanitize_json_value(val)
        lines.append(json.dumps(obj))
    return "\n".join(lines)


def merge_agents_md_skills(existing: str, skill_names: List[str], template: str) -> str:
    """Merge always-activate skill entries into an AGENTS.md file.

    `template` is the AGENTS.md snippet from the init toolkit whose first
    heading defines the section to merge into.  When the heading already
    exists in `existing`, new entries are inserted right after it (and its
    first non-heading line).  Otherwise the whole template plus entries are
    appended.

    Raises `ValueError` when `template` has no markdown heading.

    For each skill, checks if it's already listed (`skill_name` present).
    Preserves all user content.
    """
    # deduplicate while preserving order, skip already-present skills
    to_add: List[str] = []
    seen: Set[str] = set()
    for name in skill_names:
        if name in seen:
            continue
        seen.add(name)
        if ("`%s`" % name) in existing:
            continue
        to_add.append(name)

    if not to_add:
        return existing

    # parse template heading + subheading
    tpl = MarkdownDocument(template)
    tpl_heading = tpl.find_first_heading()
    if tpl_heading is None:
        raise ValueError("AGENTS.md template must contain a markdown heading")

    heading = tpl.lines[tpl_heading.line_index]
    sub_idx = tpl_heading.line_index + 1
    subheading = tpl.lines[sub_idx].strip() if sub_idx < len(tpl.lines) else None

    new_lines = ["- `%s`" % name for name in to_add]

    doc = MarkdownDocument(existing) if existing else MarkdownDocument("")
    heading_idx = doc.find_line(heading)

    if heading_idx is not None:
        insert_idx = heading_idx + 1
        if (
            subheading
            and insert_idx < len(doc.lines)
            and doc.lines[insert_idx].strip() == subheading
        ):
            insert_idx += 1
        doc.insert_lines(insert_idx, new_lines)
        return str(doc)
    else:
        section_lines = list(tpl.lines)
        section_lines.extend(new_lines)
        section_doc = MarkdownDocument("\n".join(section_lines) + "\n")
        if existing and not existing.endswith("\n"):
            doc.insert_lines(len(doc.lines), [""])
        doc.insert_md(None, MarkdownDocument("\n" + str(section_doc)))
        return str(doc)



import functools
from typing import Any
from inspect import _has_code_flag  # type: ignore[attr-defined]

try:
    from inspect import CO_GENERATOR as _CO_GENERATOR
    from inspect import CO_COROUTINE as _CO_COROUTINE
    from inspect import CO_ASYNC_GENERATOR as _CO_ASYNC_GENERATOR
except ImportError:  # pre-3.11
    _CO_GENERATOR = 0x20  # type: ignore[misc]
    _CO_COROUTINE = 0x0080  # type: ignore[misc]
    _CO_ASYNC_GENERATOR = 0x0200  # type: ignore[misc]


def _find_code_flag(obj: Any, flag: int, *, follow_wrapped: bool = True) -> bool:
    """Return True if obj (or an inner layer) has given code flag.

    The test mimics inspect.isgeneratorfunction but is more permissive:

    * It unwraps functools.partial and functools.partialmethod wrappers.
    * When follow_wrapped is True (default) it also follows the
      __wrapped__ breadcrumb that decorators created with
      functools.wraps leave behind.

    Walking stops as soon as a layer whose definition contains a yield
    statement is encountered - even if outer wrappers later hide that fact.

    Args:
        obj: The callable to inspect.
        flag: The code flag to check for.
        follow_wrapped: Follow __wrapped__ links left by decorators that
            respect the functools.wraps convention. Set to False to
            inspect only the outermost object (after stripping partial
            wrappers).

    Returns:
        True if a function definition with the specified flag was found,
        otherwise False.

    Notes:
        The check is syntactic: it looks at the code-flag on each layer.
        A regular function that merely returns a generator object will not
        satisfy the test.
    """
    while True:
        if _has_code_flag(obj, flag):
            return True

        nxt: Any = None

        # functools.partial / partialmethod wrappers
        if isinstance(obj, functools.partial):
            nxt = obj.func
        elif isinstance(obj, getattr(functools, "partialmethod", tuple())):
            nxt = obj.func
        # decorator layers exposing __wrapped__
        elif follow_wrapped and hasattr(obj, "__wrapped__"):
            nxt = obj.__wrapped__

        if nxt is None:
            # reached a layer we must trust; nothing generator-like so far.
            return False

        obj = nxt


def isgeneratorfunction(obj: Any, *, follow_wrapped: bool = True) -> bool:
    """Return True if obj (or chain of wrappers) is a generator function.

    It will follow both partial and wrapped chains and will stop on the first
    encountered generator function - contrary to `inspect` module implementation
    it checks flags on all wrappers.

    NOTE: intended to be used on user-supplied function where we do not control the wrapping

    Args:
        obj: The callable to inspect.
        follow_wrapped: Follow __wrapped__ links left by decorators that
            respect the functools.wraps convention. Set to False to
            inspect only the outermost object.

    Returns:
        True if obj is a generator function, otherwise False.
    """
    return _find_code_flag(obj, _CO_GENERATOR, follow_wrapped=follow_wrapped)


def iscoroutinefunction(obj: Any, *, follow_wrapped: bool = True) -> bool:
    """Return True if obj (or chain of wrappers) is a generator function.

    It will follow both partial and wrapped chains and will stop on the first
    encountered coroutine function - contrary to `inspect` module implementation
    it checks flags on all wrappers.

    NOTE: intended to be used on user-supplied function where we do not control the wrapping

    Args:
        obj: The callable to inspect.
        follow_wrapped: Follow __wrapped__ links left by decorators that
            respect the functools.wraps convention. Set to False to
            inspect only the outermost object.

    Returns:
        True if obj is a coroutine function, otherwise False.
    """
    return _find_code_flag(obj, _CO_COROUTINE, follow_wrapped=follow_wrapped)


def isasyncgenfunction(obj: Any, *, follow_wrapped: bool = True) -> bool:
    """Return True if obj (or chain of wrappers) is a generator function.

    It will follow both partial and wrapped chains and will stop on the first
    encountered async generator function - contrary to `inspect` module implementation
    it checks flags on all wrappers.

    NOTE: intended to be used on user-supplied function where we do not control the wrapping

    Args:
        obj: The callable to inspect.
        follow_wrapped: Follow __wrapped__ links left by decorators that
            respect the functools.wraps convention. Set to False to
            inspect only the outermost object.

    Returns:
        True if obj is an async generator function, otherwise False.
    """
    return _find_code_flag(obj, _CO_ASYNC_GENERATOR, follow_wrapped=follow_wrapped)



from typing import Any

from dlt.extract.exceptions import PipeException
from dlt.common.typing import TDataItem


class IncrementalCursorPathMissing(PipeException):
    def __init__(
        self, pipe_name: str, json_path: str, item: TDataItem = None, msg: str = None
    ) -> None:
        self.json_path = json_path
        self.item = item
        msg = (
            msg
            or f"Cursor element with JSON path `{json_path}` was not found in extracted data item. All data items must contain this path. Use the same names of fields as in your JSON document because they can be different from the names you see in database."
        )
        super().__init__(pipe_name, msg)


class IncrementalCursorPathHasValueNone(PipeException):
    def __init__(
        self, pipe_name: str, json_path: str, item: TDataItem = None, msg: str = None
    ) -> None:
        self.json_path = json_path
        self.item = item
        msg = (
            msg
            or f"Cursor element with JSON path `{json_path}` has the value `None` in extracted data item. All data items must contain a value != None. Construct the incremental with `on_cursor_value_missing='include'` if you want to include such rows"
        )
        super().__init__(pipe_name, msg)


class IncrementalCursorInvalidCoercion(PipeException):
    def __init__(
        self,
        pipe_name: str,
        cursor_path: str,
        cursor_value: TDataItem,
        cursor_value_type: str,
        item: TDataItem,
        item_type: Any,
        details: str,
    ) -> None:
        self.cursor_path = cursor_path
        self.cursor_value = cursor_value
        self.cursor_value_type = cursor_value_type
        self.item = item
        msg = (
            f"Could not coerce `{cursor_value_type}` with value `{cursor_value}` and type"
            f" `{type(cursor_value)}` to actual data item `{item}` at path `{cursor_path}` with"
            f" type `{item_type}`: {details}. You need to use different data type for"
            f" `{cursor_value_type}` or cast your data ie. by using `.add_map()` on this resource."
        )
        super().__init__(pipe_name, msg)


class IncrementalPrimaryKeyMissing(PipeException):
    def __init__(self, pipe_name: str, primary_key_column: str, item: TDataItem) -> None:
        self.primary_key_column = primary_key_column
        self.item = item
        msg = (
            f"Primary key column `{primary_key_column}` was not found in extracted data item. All"
            " data items must contain this column. Use the same names of fields as in your JSON"
            " document."
        )
        super().__init__(pipe_name, msg)



import logging
from typing import (
    Iterator,
    Optional,
    List,
    Dict,
    Any,
    TypeVar,
    Iterable,
    cast,
    Callable,
)
import copy
from urllib.parse import urlparse
from requests import Session as BaseSession  # noqa: I251
from requests import Response, Request, HTTPError
from requests.auth import AuthBase

from dlt.common import jsonpath, logger
from dlt.common.configuration import resolve_configuration
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration

from .typing import HTTPMethodBasic, HTTPMethod, Hooks
from .paginators import BasePaginator
from .detector import PaginatorFactory, find_response_page_data
from .exceptions import IgnoreResponseException, PaginatorNotFound
from .redaction import sanitize_url
from .utils import join_url


_T = TypeVar("_T")


class PageData(List[_T]):
    """A list of elements in a single page of results with attached request context.

    The context allows to inspect the response, paginator and authenticator, modify the request
    """

    def __init__(
        self,
        __iterable: Iterable[_T],
        request: Request,
        response: Response,
        paginator: BasePaginator,
        auth: AuthBase,
    ):
        super().__init__(__iterable)
        self.request = request
        self.response = response
        self.paginator = paginator
        self.auth = auth


class RESTClient:
    """A generic REST client for making requests to an API with support for
    pagination and authentication.

    Args:
        base_url (str): The base URL of the API to make requests to.
        headers (Optional[Dict[str, str]]): Default headers to include in all requests.
        auth (Optional[AuthBase]): Authentication configuration for all requests.
        paginator (Optional[BasePaginator]): Default paginator for handling paginated responses.
        data_selector (Optional[jsonpath.TJsonPath]): JSONPath selector for extracting data from responses.
            Only used in `paginate`.
        session (BaseSession): HTTP session for making requests.
        paginator_factory (Optional[PaginatorFactory]): Factory for creating paginator instances,
            used for detecting paginators.
        config (Optional[RuntimeConfiguration]): Runtime configuration for HTTP error handling.
            If not provided, will resolve from the global configuration.
    """

    def __init__(
        self,
        base_url: str,
        headers: Optional[Dict[str, str]] = None,
        auth: Optional[AuthBase] = None,
        paginator: Optional[BasePaginator] = None,
        data_selector: Optional[jsonpath.TJsonPath] = None,
        session: BaseSession = None,
        paginator_factory: Optional[PaginatorFactory] = None,
        config: Optional[RuntimeConfiguration] = None,
    ) -> None:
        self.base_url = base_url
        self.headers = headers
        self.auth = auth
        self._config = config or resolve_configuration(RuntimeConfiguration())

        if session:
            # If the `session` is provided (for example, an instance of
            # dlt.sources.helpers.requests.session.Session), warn if
            # it has raise_for_status=True by default
            self.session = _warn_if_raise_for_status_and_return(session)
        else:
            # Otherwise, create a new Client with disabled raise_for_status
            # to allow for custom error handling in the hooks
            from dlt.sources.helpers.requests.retry import Client

            self.session = Client(raise_for_status=False).session

        self.paginator = paginator
        self.pagination_factory = paginator_factory or PaginatorFactory()

        self.data_selector = data_selector

    def _create_request(
        self,
        path_or_url: str,
        method: HTTPMethod,
        headers: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Any] = None,
        auth: Optional[AuthBase] = None,
        hooks: Optional[Hooks] = None,
    ) -> Request:
        if json and data:
            raise ValueError(
                "Cannot use both 'json' and 'data' parameters simultaneously. "
                "Use 'json' for JSON payloads or 'data' for form-encoded/raw data."
            )
        parsed_url = urlparse(path_or_url)
        if parsed_url.scheme in ("http", "https"):
            url = path_or_url
        else:
            url = join_url(self.base_url, path_or_url)

        request_headers = (self.headers or {}) | (headers or {})

        return Request(
            method=method,
            url=url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            auth=auth or self.auth,
            hooks=hooks,
        )

    def _send_request(self, request: Request, **kwargs: Any) -> Response:
        prepared_request = self.session.prepare_request(request)

        self._log_request(request, prepared_request.url)

        send_kwargs = self.session.merge_environment_settings(
            prepared_request.url,
            kwargs.pop("proxies", {}),
            kwargs.pop("stream", None),
            kwargs.pop("verify", None),
            kwargs.pop("cert", None),
        )

        send_kwargs.update(**kwargs)  #  type: ignore[call-arg]
        return self.session.send(prepared_request, **send_kwargs)

    def _log_request(self, request: Request, prepared_url: str) -> None:
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Making {request.method.upper()} request to {request.url}"
                f" with params={request.params}, json={request.json},"
                f" data={request.data}, headers={request.headers}"
            )
        else:
            sanitized_url = sanitize_url(prepared_url)
            logger.info(f"Making {request.method.upper()} request to {sanitized_url}")

    def request(self, path: str = "", method: HTTPMethod = "GET", **kwargs: Any) -> Response:
        prepared_request = self._create_request(
            path_or_url=path,
            method=method,
            headers=kwargs.pop("headers", None),
            params=kwargs.pop("params", None),
            json=kwargs.pop("json", None),
            data=kwargs.pop("data", None),
            auth=kwargs.pop("auth", None),
            hooks=kwargs.pop("hooks", None),
        )
        return self._send_request(prepared_request, **kwargs)

    def get(self, path: str, params: Optional[Dict[str, Any]] = None, **kwargs: Any) -> Response:
        return self.request(path, method="GET", params=params, **kwargs)

    def post(
        self,
        path: str,
        json: Optional[Dict[str, Any]] = None,
        *,
        data: Optional[Any] = None,
        **kwargs: Any,
    ) -> Response:
        return self.request(path, method="POST", json=json, data=data, **kwargs)

    def paginate(
        self,
        path: str = "",
        method: HTTPMethodBasic = "GET",
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        auth: Optional[AuthBase] = None,
        paginator: Optional[BasePaginator] = None,
        data_selector: Optional[jsonpath.TJsonPath] = None,
        hooks: Optional[Hooks] = None,
        headers: Optional[Dict[str, Any]] = None,
        *,
        data: Optional[Any] = None,
        **kwargs: Any,
    ) -> Iterator[PageData[Any]]:
        """Iterates over paginated API responses, yielding pages of data.

        Args:
            path (str): Endpoint path for the request, relative to `base_url`.
                Can also be a fully qualified URL; if starting with http(s) it will
                be used instead of the base_url + path.
            method (HTTPMethodBasic): HTTP method for the request, defaults to 'get'.
            headers (Optional[Dict[str, Any]]): Headers for the request.
                Overwrites default headers.
            params (Optional[Dict[str, Any]]): URL parameters for the request.
            json (Optional[Dict[str, Any]]): JSON payload for the request.
            data (Optional[Any]): Data to send in the body of the request. Can be a dictionary
                (form-encoded), bytes, string, or file-like object. Mutually exclusive with json.
            auth (Optional[AuthBase): Authentication configuration for the request.
            paginator (Optional[BasePaginator]): Paginator instance for handling
                pagination logic.
            data_selector (Optional[jsonpath.TJsonPath]): JSONPath selector for
                extracting data from the response.
            hooks (Optional[Hooks]): Hooks to modify request/response objects. Note that
                when hooks are not provided, the default behavior is to raise an exception
                on error status codes.
            **kwargs (Any): Optional arguments to that the Request library accepts, such as
                `stream`, `verify`, `proxies`, `cert`, `timeout`, and `allow_redirects`.

        Yields:
            PageData[Any]: A page of data from the paginated API response, along with request
                and response context.

        Raises:
            HTTPError: If the response status code is not a success code. This is raised
                by default when hooks are not provided.

        Example:
            >>> client = RESTClient(base_url="https://api.example.com")
            >>> for page in client.paginate("/search", method="post", json={"query": "foo"}):
            >>>     print(page)
        """
        paginator = paginator if paginator else copy.deepcopy(self.paginator)
        auth = auth or self.auth
        data_selector = data_selector or self.data_selector
        hooks = hooks or {}

        # Add the raise_for_status hook to ensure an exception is raised on
        # HTTP error status codes. This is a fallback to handle errors
        # unless explicitly overridden in the provided hooks.
        if "response" not in hooks:
            hooks["response"] = [self._create_response_handler()]

        request = self._create_request(
            path_or_url=path,
            headers=headers,
            method=method,
            params=params,
            json=json,
            data=data,
            auth=auth,
            hooks=hooks,
        )

        if paginator:
            paginator.init_request(request)

        while True:
            try:
                response = self._send_request(request, **kwargs)
            except IgnoreResponseException:
                break

            if not data_selector:
                data_selector = self.detect_data_selector(response)
            data = self.extract_response(response, data_selector)

            if paginator is None:
                paginator = self.detect_paginator(response, data)
            paginator.update_state(response, data)
            paginator.update_request(request)

            # yield data with context
            yield PageData(data, request=request, response=response, paginator=paginator, auth=auth)

            if not paginator.has_next_page:
                logger.info(f"Paginator {str(paginator)} does not have more pages")
                break

    def extract_response(self, response: Response, data_selector: jsonpath.TJsonPath) -> List[Any]:
        # we should compile data_selector
        data: Any = jsonpath.find_values(data_selector, response.json())

        # extract if single item selected
        data = data[0] if isinstance(data, list) and len(data) == 1 else data
        if isinstance(data, list):
            length_info = f" with length {len(data)}"
        else:
            length_info = ""
        logger.info(
            f"Extracted data of type {type(data).__name__} from path {data_selector}{length_info}"
        )
        if data is None:
            logger.warning(
                "Extracted data is None for path %s, returning empty list", data_selector
            )
            return []
        # wrap single pages into lists
        if not isinstance(data, list):
            data = [data]
        return cast(List[Any], data)

    def detect_data_selector(self, response: Response) -> str:
        """Detects a path to page data in `response`. If there's no
           paging detected, returns "$" which will select full response

        Returns:
            str: a json path to the page data.
        """
        path, data = find_response_page_data(response.json())
        data_selector = ".".join(path)
        # if list is detected, it is probably a paged data
        if isinstance(data, list):
            logger.info(
                f"Detected page data at path: '{data_selector}' type: list length: {len(data)}"
            )
        else:
            logger.info(f"Detected single page data at path: '{path}' type: {type(data).__name__}")
        return data_selector

    def _create_response_handler(self) -> Callable[[Response, Any, Any], None]:
        def handler(response: Response, *args: Any, **kwargs: Any) -> None:
            return _dlt_raise_for_status(
                response,
                self._config.http_show_error_body,
                self._config.http_max_error_body_length,
                *args,
                **kwargs,
            )

        return handler

    def detect_paginator(self, response: Response, data: Any) -> BasePaginator:
        """Detects a paginator for the response and returns it.

        Args:
            response (Response): The response to detect the paginator for.
            data_selector (data_selector): Path to paginated data or $ if paginated data not detected

        Returns:
            BasePaginator: The paginator instance that was detected.
        """
        paginator, score = self.pagination_factory.create_paginator(response)
        if paginator is None:
            raise PaginatorNotFound(
                f"No suitable `paginator` found for the response at `{response.url}`"
            )
        if score == 1.0:
            logger.info(f"Detected `paginator`: `{paginator}`")
        elif score == 0.0:
            if isinstance(data, list):
                logger.warning(
                    f"Fallback `paginator` used: `{paginator}`. Please provide paginator manually."
                )
            else:
                logger.info(
                    f"Fallback `paginator` used: `{paginator}` to handle not paginated response."
                )
        else:
            logger.warning(
                "Please verify the `paginator` settings. We strongly suggest to use explicit"
                " instance of the paginator as some settings may not be guessed correctly."
            )
        return paginator


def _dlt_raise_for_status(
    response: Response, show_error_body: bool, max_error_body_length: int, *args: Any, **kwargs: Any
) -> None:
    """Custom raise_for_status that sanitizes URLs and includes response body.

    Args:
        response: The response object to check
        show_error_body: Whether to show the error body
        max_error_body_length: The maximum length of the error body to show

    Raises:
        HTTPError: If the response status code indicates an error
    """
    if response.status_code >= 400:
        safe_url = sanitize_url(response.url)

        body_text = ""

        if show_error_body and response.text:
            body_text = response.text
            if len(body_text) > max_error_body_length:
                body_text = body_text[:max_error_body_length] + "… (truncated)"

        reason = _decode_reason(response.reason)
        error_type_string = "Client" if response.status_code < 500 else "Server"

        msg = f"{response.status_code} {error_type_string} Error: {reason} for url: {safe_url}"
        if body_text:
            msg += f"\nResponse: {body_text}"

        raise HTTPError(msg, response=response)


def _decode_reason(reason: Any) -> Any:
    if isinstance(reason, bytes):
        try:
            return reason.decode("utf-8")
        except UnicodeDecodeError:
            return reason.decode("iso-8859-1")
    return reason


def raise_for_status(response: Response, *args: Any, **kwargs: Any) -> None:
    response.raise_for_status()


def _warn_if_raise_for_status_and_return(session: BaseSession) -> BaseSession:
    """A generic function to warn if the session has raise_for_status enabled."""
    if getattr(session, "raise_for_status", False):
        logger.warning(
            "The session provided has raise_for_status enabled. This may cause unexpected behavior."
        )
    return session



import typing as t

from dlt.common.destination import Destination, DestinationCapabilitiesContext

from dlt.common.destination.capabilities import TLoaderFileFormat
from dlt.common.normalizers.naming.naming import NamingConvention
from dlt.destinations.impl.dummy.configuration import (
    DummyClientConfiguration,
    DummyClientCredentials,
)
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE

if t.TYPE_CHECKING:
    from dlt.destinations.impl.dummy.dummy import DummyClient


class dummy(Destination[DummyClientConfiguration, "DummyClient"]):
    spec = DummyClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_staging_file_format = None
        caps.has_case_sensitive_identifiers = True
        caps.max_identifier_length = 127
        caps.max_column_identifier_length = 127
        caps.max_query_length = 8 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 65536
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = False
        caps.supported_merge_strategies = ["delete-insert", "upsert"]

        return caps

    @property
    def client_class(self) -> t.Type["DummyClient"]:
        from dlt.destinations.impl.dummy.dummy import DummyClient

        return DummyClient

    def __init__(
        self,
        credentials: DummyClientCredentials = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(
            credentials=credentials,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: DummyClientConfiguration,
        naming: t.Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        caps = super().adjust_capabilities(caps, config, naming)
        additional_formats: t.List[TLoaderFileFormat] = (
            ["reference"]
            if (config.create_followup_jobs or config.create_followup_table_chain_reference_jobs)
            else []
        )
        caps.preferred_loader_file_format = config.loader_file_format
        caps.supported_loader_file_formats = additional_formats + [config.loader_file_format]
        caps.supported_staging_file_formats = additional_formats + [config.loader_file_format]
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)

        return caps


dummy.register()



"""Pipeline execution timeline visualization and load package status badges."""

from typing import Dict, List, Optional, Tuple, NamedTuple, Union, cast, get_args

from dlt.common.pendulum import datetime

import marimo as mo

from dlt.common.pendulum import pendulum
from dlt.common.pipeline import LoadInfo
from dlt.common.storages import LoadPackageInfo
from dlt.common.storages.load_package import PackageStorage

from dlt._workspace.helpers.dashboard.const import (
    LOAD_PACKAGE_STATUS_COLORS,
    PENDING_LOAD_STATUSES,
    PIPELINE_RUN_STEP_COLORS,
    TPipelineRunStatus,
    TVisualPipelineStep,
    VISUAL_PIPELINE_STEPS,
)
from dlt._workspace.helpers.dashboard.utils.formatters import format_duration
from dlt._workspace.helpers.dashboard.utils.ui import dlt_table
from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace


class PipelineStepData(NamedTuple):
    step: TVisualPipelineStep
    duration_ms: float
    failed: bool


def badge_html(text: str, color: str, size: str = "small") -> str:
    """Build a colored badge HTML element."""
    return f'<div class="status-badge status-badge-{color}"><{size}>{text}</{size}></div>'


def migration_badge(count: int) -> str:
    """Build migration badge HTML using CSS classes."""
    if count == 0:
        return ""
    return badge_html(f"{count} dataset migration(s)", "yellow", "strong")


def status_badge(status: TPipelineRunStatus) -> str:
    """Build status badge HTML using CSS classes."""
    return badge_html(status, "green" if status == "succeeded" else "red", "strong")


def pipeline_execution_html(
    transaction_id: str,
    status: TPipelineRunStatus,
    steps_data: List[PipelineStepData],
    migrations_count: int = 0,
    finished_at: Optional[datetime] = None,
) -> mo.Html:
    """Build an HTML visualization for a pipeline execution using CSS classes."""
    total_ms = sum(step.duration_ms for step in steps_data)
    if total_ms == 0:
        total_ms = 1.0  # avoid division by zero when all steps complete instantly
    last = len(steps_data) - 1

    # Build the general info of the execution
    relative_time = ""
    if finished_at:
        time_ago = pendulum.instance(finished_at).diff_for_humans()
        relative_time = f"<div><small>Executed: <strong>{time_ago}</strong></small></div>"

    general_info = f"""
    <div><small>Last execution ID: <strong>{transaction_id[:8]}</strong></small></div>
    <div><small>Total time: <strong>{format_duration(total_ms)}</strong></small></div>
    {relative_time}
    """

    # Build the pipeline execution timeline bar and labels
    segments, labels = [], []
    for i, step in enumerate(steps_data):
        percentage = step.duration_ms / total_ms * 100
        color = PIPELINE_RUN_STEP_COLORS.get(step.step)
        radius = (
            "6px"
            if i == 0 and i == last
            else "6px 0 0 6px" if i == 0 else "0 6px 6px 0" if i == last else "0"
        )
        segments.append(
            '<div class="pipeline-execution-timeline-segment" '
            f'style="width:{percentage}%;background-color:{color};border-radius:{radius};"></div>'
        )
        labels.append(
            f'<span><span style="color:{color};">●</span> '
            f"{step.step.capitalize()} <strong>{format_duration(step.duration_ms)}</strong> </span>"
        )

    html = f"""
    <div class="pipeline-execution-container">
        <!-- Main 3-column flex container -->
        <div class="pipeline-execution-layout">

            <!-- LEFT COLUMN: Run ID, Total time -->
            <div class="pipeline-execution-info paragraph">
                {general_info}
            </div>

            <!-- CENTER COLUMN: Timeline bar and legend -->
            <div class="pipeline-execution-timeline">
                <div class="pipeline-execution-timeline-bar">
                    {''.join(segments)}
                </div>
                <div class="pipeline-execution-labels">
                    {''.join(labels)}
                </div>
            </div>

            <!-- RIGHT COLUMN: Status badges -->
            <div class="pipeline-execution-badges">
                {migration_badge(migrations_count)}
                {status_badge(status)}
            </div>
        </div>
    </div>
    """
    return mo.Html(html)


def get_steps_data_and_status(
    trace_steps: List[PipelineStepTrace],
) -> Tuple[List[PipelineStepData], TPipelineRunStatus]:
    """Get trace steps data and the status of the corresponding pipeline execution."""
    steps_data: List[PipelineStepData] = []
    any_step_failed: bool = False

    for step in trace_steps:
        if step.step_exception is not None:
            any_step_failed = True

        if step.step not in get_args(TVisualPipelineStep) or not step.finished_at:
            continue

        duration_ms = (step.finished_at - step.started_at).total_seconds() * 1000
        steps_data.append(
            PipelineStepData(
                step=cast(TVisualPipelineStep, step.step),
                duration_ms=duration_ms,
                failed=step.step_exception is not None,
            )
        )
    status: TPipelineRunStatus = "failed" if any_step_failed else "succeeded"
    return steps_data, status


def get_migrations_count(last_load_info: LoadInfo) -> int:
    """Count the number of unique migrations (schema versions) from load packages."""
    migrations_count: int = 0
    seen_schema_hashes = set()
    for package in last_load_info.load_packages:
        if len(package.schema_update) > 0:
            if package.schema_hash not in seen_schema_hashes:
                migrations_count += 1
                seen_schema_hashes.add(package.schema_hash)
    return migrations_count


def pipeline_execution_visualization(trace: PipelineTrace) -> Optional[mo.Html]:
    """Create a visual timeline of pipeline run showing extract, normalize and load steps."""
    steps_data, status = get_steps_data_and_status(trace.steps)
    migrations_count = get_migrations_count(trace.last_load_info) if trace.last_load_info else 0

    return pipeline_execution_html(
        trace.transaction_id,
        status,
        steps_data,
        migrations_count,
        trace.finished_at,
    )


def collect_load_packages_from_trace(
    trace: PipelineTrace,
) -> List[LoadPackageInfo]:
    """Collect all unique load packages from all steps."""
    packages_by_load_id: Dict[str, LoadPackageInfo] = {}

    for step in trace.steps:
        if step.step in VISUAL_PIPELINE_STEPS and step.step_info and step.step_info.load_packages:
            for package in step.step_info.load_packages:
                packages_by_load_id[package.load_id] = package

    return list(packages_by_load_id.values())


def load_package_status_labels(trace: PipelineTrace) -> mo.ui.table:
    """Build a table of load package status badges for each package in the trace.

    For each package, determines its visual status badge based on whether the
    package is partially loaded, pending, or in a final state.
    """
    packages = collect_load_packages_from_trace(trace)
    result: List[Dict[str, Union[str, mo.Html]]] = []

    for package in packages:
        is_partial = PackageStorage.is_package_partially_loaded(package)
        badge_color_key = "red" if is_partial else LOAD_PACKAGE_STATUS_COLORS.get(package.state)
        if is_partial:
            badge_text = f"partially {package.state}"
        elif package.state in PENDING_LOAD_STATUSES:
            badge_text = PENDING_LOAD_STATUSES.get(package.state)
        elif package.state == "new":
            badge_text = "discarded"
        else:
            badge_text = package.state

        status_html = badge_html(badge_text, badge_color_key, "strong")
        result.append(
            {
                "load_id": package.load_id,
                "status": mo.Html(status_html),
            }
        )

    return dlt_table(
        result,
        pagination=True,
        show_download=False,
        freeze_column=None,
    )



from typing import Literal, Optional

from dlt.common.schema.typing import (
    _TTableSchemaBase,
    TWriteDisposition,
    TTableReferenceParam,
)


class PreparedTableSchema(_TTableSchemaBase, total=False):
    """Table schema with all hints prepared to be loaded"""

    write_disposition: TWriteDisposition
    references: Optional[TTableReferenceParam]
    _x_prepared: bool  # needed for the type checker



from typing import Any, TYPE_CHECKING, Tuple, List
from packaging.version import Version
import duckdb

from dlt.common import logger
from dlt.common.destination.exceptions import DestinationUndefinedEntity
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.schema.utils import is_nullable_column
from dlt.common.storages.configuration import FileSystemCredentials

from dlt.common.typing import TLoaderFileFormat
from dlt.destinations.sql_client import raise_database_error
from dlt.destinations.impl.duckdb.sql_client import WithTableScanners
from dlt.destinations.impl.duckdb.factory import DuckDbCredentials

from dlt.destinations.utils import is_compression_disabled
from dlt.destinations.path_utils import get_file_format_and_compression

SUPPORTED_PROTOCOLS = ["gs", "gcs", "s3", "file", "memory", "az", "abfss", "hf"]

if TYPE_CHECKING:
    from dlt.destinations.impl.filesystem.filesystem import FilesystemClient
else:
    FilesystemClient = Any


class FilesystemSqlClient(WithTableScanners):
    def __init__(
        self,
        remote_client: FilesystemClient,
        dataset_name: str,
        cache_db: DuckDbCredentials = None,
        persist_secrets: bool = False,
    ) -> None:
        if remote_client.config.protocol not in SUPPORTED_PROTOCOLS:
            raise NotImplementedError(
                f"Received invalid value `protocol={remote_client.config.protocol}` for"
                f" `FilesystemSqlClient`. Valid values are: {SUPPORTED_PROTOCOLS}"
            )

        super().__init__(remote_client, dataset_name, cache_db, persist_secrets=persist_secrets)
        self.remote_client: FilesystemClient = remote_client
        self.is_abfss = self.remote_client.config.protocol == "abfss"
        self.iceberg_initialized = False
        if self.is_abfss:
            self._global_config["azure_transport_option_type"] = "curl"

    def can_create_view(self, table_schema: PreparedTableSchema) -> bool:
        if table_schema.get("table_format") in ("delta", "iceberg"):
            return True
        # checking file type is expensive so we optimistically allow to create view and prune later
        return True

    def get_file_format_and_files(
        self, table_schema: PreparedTableSchema
    ) -> Tuple[str, List[str], bool]:
        table_name = table_schema["name"]
        files = self.remote_client.list_table_files(table_name)
        if len(files) == 0:
            raise DestinationUndefinedEntity(table_name)
        file_format, is_compressed = get_file_format_and_compression(files[0])
        return file_format, files, is_compressed

    def create_secret(
        self,
        scope: str,
        credentials: FileSystemCredentials,
        secret_name: str = None,
        persist_secrets: bool = False,
    ) -> bool:
        protocol = self.remote_client.config.protocol
        if protocol == "file":
            return True
        if not super().create_secret(
            scope, credentials, secret_name=secret_name, persist_secrets=persist_secrets
        ):
            # native google storage implementation is not supported..
            if protocol in ["gs", "gcs"]:
                logger.warn(
                    "For gs/gcs access via duckdb please use the gs/gcs s3 compatibility layer if"
                    " possible (not supported when using `iceberg` table format). Falling back to"
                    " fsspec."
                )
                self._conn.register_filesystem(self.remote_client.fs_client)
            # for memory we also need to register filesystem
            elif protocol == "memory":
                self._conn.register_filesystem(self.remote_client.fs_client)
            elif protocol == "file":
                # authentication for local filesystem not needed
                pass
            else:
                raise ValueError(f"Cannot create secret or register filesystem for `{protocol=:}`")

        return True

    def open_connection(self) -> duckdb.DuckDBPyConnection:
        with self.credentials.conn_pool._conn_lock:
            first_connection = self.credentials.conn_pool.never_borrowed
            super().open_connection()

        if first_connection:
            # TODO: we need to frontload the httpfs extension for abfss for some reason
            if self.is_abfss:
                self._conn.sql("INSTALL httpfs; LOAD httpfs")

            # create single authentication for the whole client
            self.create_secret(
                self.remote_client.config.bucket_url,
                self.remote_client.config.credentials,
                persist_secrets=self.persist_secrets,
            )
        return self._conn

    def should_replace_view(self, view_name: str, table_schema: PreparedTableSchema) -> bool:
        if self.remote_client.config.always_refresh_views:
            table_format = table_schema.get("table_format")
            if table_format == "delta":
                # delta will auto refresh
                return False
        return self.remote_client.config.always_refresh_views

    @raise_database_error
    def create_view(self, view_name: str, table_schema: PreparedTableSchema) -> None:
        # NOTE: data freshness
        # iceberg - currently we glob the most recent snapshot (via built in duckdb mechanism) so data is fresh
        #           (but not very efficient)
        # delta - newest version is always read
        # files - newest files
        table_name = table_schema["name"]
        table_format = table_schema.get("table_format")
        protocol = self.remote_client.config.protocol
        table_location = self.remote_client.get_open_table_location(table_format, table_name)

        dlt_table_names = self.remote_client.schema.dlt_table_names()

        def _escape_column_name(col_name: str) -> str:
            col_name = self.escape_column_name(col_name)
            # dlt tables are stored as json and never normalized
            if table_name in dlt_table_names:
                col_name = col_name.lower()
            return col_name

        # get columns to select from table schema
        columns = [_escape_column_name(c) for c in self.schema.get_table_columns(table_name).keys()]

        # create from statement
        from_statement = ""
        if table_format == "delta":
            table_location = table_location.rstrip("/")
            from_statement = f"delta_scan('{table_location}')"
        elif table_format == "iceberg":
            table_location = table_location.rstrip("/")
            if not self.iceberg_initialized:
                self._setup_iceberg(self._conn)
                self.iceberg_initialized = True

            from dlt.common.libs.pyiceberg import get_last_metadata_file

            metadata_path = f"{table_location}/metadata"
            last_metadata_file = get_last_metadata_file(
                metadata_path, self.remote_client.fs_client, self.remote_client.config
            )
            if ".gz." in last_metadata_file:
                compression = ", metadata_compression_codec = 'gzip'"
            else:
                compression = ""

            if Version(duckdb.__version__) > Version("1.3.0"):
                scanner_options = "union_by_name=true"
            else:
                scanner_options = "skip_schema_inference=false"

            from_statement = f"iceberg_scan('{last_metadata_file}'{compression}, {scanner_options})"
            # TODO: on duckdb > 1.2.1 register self.remote_client.fs_client as abfss fsspec filesystem
            #   this will enable iceberg but with lower performance
        else:
            # get file format and list of table files
            # NOTE: this does not support cases where table contains many different file formats
            # NOTE: since we must list all the files anyway we just pass them to duckdb without further globbing
            #   list is in the memory already and query size in duckdb is very large
            first_file_type, files, _ = self.get_file_format_and_files(table_schema)
            if protocol == "file":
                resolved_files_string = ",".join(map(lambda f: f"'{f}'", files))
            else:
                resolved_files_string = ",".join(map(lambda f: f"'{protocol}://{f}'", files))

            # NOTE: duckdb automatically handles compression based on the .gz extension
            # so we don't need to specify it
            if first_file_type == "parquet":
                from_statement = f"read_parquet([{resolved_files_string}], union_by_name=true)"
            elif first_file_type in ("jsonl", "csv"):
                # build columns definition
                type_mapper = self.capabilities.get_type_mapper()
                columns_defs = self.schema.get_table_columns(table_name).values()
                column_types = ",".join(
                    map(
                        lambda c: (
                            f'{_escape_column_name(c["name"])}:'
                            f' "{type_mapper.to_destination_type(c, table_schema)}"'
                        ),
                        columns_defs,
                    )
                )
                if first_file_type == "jsonl":
                    # swap binary types
                    for idx, column_def in enumerate(columns_defs):
                        if column_def["data_type"] == "binary":
                            columns[idx] = f"from_base64(decode({columns[idx]})) as {columns[idx]}"
                    from_statement = (
                        f"read_json([{resolved_files_string}], columns = {{{column_types}}})"
                    )
                if first_file_type == "csv":
                    # TODO: use default csv_format config to set params below
                    not_null_columns = [
                        _escape_column_name(c["name"])
                        for c in columns_defs
                        if not is_nullable_column(c)
                    ]
                    if not_null_columns:
                        force_not_null = f"force_not_null=[{','.join(not_null_columns)}],"
                    else:
                        force_not_null = ""
                    # use `types` (and not `columns`) options below. columns does not
                    # work with multiple csv files with evolving schemas. they disable
                    # autodetect and lock schema for all files
                    from_statement = (
                        f"read_csv([{resolved_files_string}],{force_not_null} union_by_name=true,header=true,null_padding=true,types="
                        f" {{{column_types}}})"
                    )

                # if the dataset is a legacy version where .gz is not added by default,
                # we need to check configs
                if (
                    table_name not in dlt_table_names
                    and self.remote_client.storage_versions[0] == 1
                    and not is_compression_disabled()
                ):
                    from_statement = from_statement[:-1] + ", compression = 'gzip')"

            else:
                # we skipped checking file type in can_create_view to not repeat globs which are expensive
                # so we skip here.
                return
                # raise NotImplementedError(
                #     f"Unknown filetype {first_file_type} for table {table_name}. Currently only"
                #     " jsonl and parquet files as well as delta and iceberg tables are"
                #     " supported."
                # )

        # create table
        view_name = self.make_qualified_table_name(view_name)
        create_table_sql_base = (
            f"CREATE OR REPLACE VIEW {view_name} AS SELECT {', '.join(columns)} FROM"
            f" {from_statement}"
        )
        self._conn.execute(create_table_sql_base)



from typing import Any

from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import TColumnNames
from dlt.extract import DltResource
from dlt.destinations.utils import get_resource_for_adapter

VECTORIZE_HINT = "x-qdrant-embed"


def qdrant_adapter(
    data: Any,
    embed: TColumnNames = None,
) -> DltResource:
    """Prepares data for the Qdrant destination by specifying which columns
    should be embedded.

    Args:
        data (Any): The data to be transformed. It can be raw data or an instance
            of DltResource. If raw data, the function wraps it into a DltResource
            object.
        embed (TColumnNames, optional): Specifies columns to generate embeddings for.
            Can be a single column name as a string or a list of column names.

    Returns:
        DltResource: A resource with applied qdrant-specific hints.

    Raises:
        ValueError: If input for `embed` invalid or empty.

    Examples:
        >>> data = [{"name": "Anush", "description": "Integrations Hacker"}]
        >>> qdrant_adapter(data, embed="description")
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)

    column_hints: TTableSchemaColumns = None

    if embed:
        if isinstance(embed, str):
            embed = [embed]
        if not isinstance(embed, list):
            raise ValueError(
                "`embed` must be a list of column names or a single column name as a string"
            )

        column_hints = {}
        for column_name in embed:
            column_hints[column_name] = {
                "name": column_name,
                VECTORIZE_HINT: True,  # type: ignore
            }

    if not column_hints:
        raise ValueError("A value for `embed` must be specified.")
    else:
        resource.apply_hints(columns=column_hints)

    return resource



from typing import Dict, Any, List
from abc import ABC, abstractmethod

from dlt.common import logger
from dlt.common.metrics import DataWriterMetrics
from dlt.common.schema import TTableSchemaColumns
from dlt.common.typing import TDataItems
from dlt.common.data_writers import (
    BufferedDataWriter,
    DataWriter,
    FileWriterSpec,
)


class DataItemStorage(ABC):
    def __init__(self, writer_spec: FileWriterSpec, *args: Any) -> None:
        self.writer_spec = writer_spec
        self.writer_cls = DataWriter.writer_class_from_spec(writer_spec)
        self.buffered_writers: Dict[str, BufferedDataWriter[DataWriter]] = {}
        super().__init__(*args)

    def _get_writer(
        self, load_id: str, schema_name: str, table_name: str
    ) -> BufferedDataWriter[DataWriter]:
        # unique writer id
        writer_id = f"{load_id}.{schema_name}.{table_name}"
        writer = self.buffered_writers.get(writer_id, None)
        if not writer:
            # assign a writer for each table
            kwargs = {}
            if self.writer_spec.file_max_items:
                kwargs["file_max_items"] = self.writer_spec.file_max_items
            path = self._get_data_item_path_template(load_id, schema_name, table_name)
            writer = BufferedDataWriter(self.writer_spec, path, **kwargs)
            self.buffered_writers[writer_id] = writer
        return writer

    def write_data_item(
        self,
        load_id: str,
        schema_name: str,
        table_name: str,
        item: TDataItems,
        columns: TTableSchemaColumns,
    ) -> int:
        writer = self._get_writer(load_id, schema_name, table_name)
        # write item(s)
        return writer.write_data_item(item, columns)

    def write_empty_items_file(
        self, load_id: str, schema_name: str, table_name: str, columns: TTableSchemaColumns
    ) -> DataWriterMetrics:
        """Writes empty file: only header and footer without actual items. Closed the
        empty file and returns metrics. Mind that header and footer will be written."""
        writer = self._get_writer(load_id, schema_name, table_name)
        return writer.write_empty_file(columns)

    def import_items_file(
        self,
        load_id: str,
        schema_name: str,
        table_name: str,
        file_path: str,
        metrics: DataWriterMetrics,
        with_extension: str = None,
    ) -> DataWriterMetrics:
        """Import a file from `file_path` into items storage under a new file name. Does not check
        the imported file format. Uses counts from `metrics` as a base. Logically closes the imported file

        The preferred import method is a hard link to avoid copying the data. If current filesystem does not
        support it, a regular copy is used.

        Alternative extension may be provided via `with_extension` so various file formats may be imported into the same folder.
        """
        writer = self._get_writer(load_id, schema_name, table_name)
        return writer.import_file(file_path, metrics, with_extension)

    def close_writers(self, load_id: str, skip_flush: bool = False) -> None:
        """Flush, write footers (skip_flush), write metrics and close files in all
        writers belonging to `load_id` package
        """
        for name, writer in self.buffered_writers.items():
            if name.startswith(load_id) and not writer.closed:
                logger.debug(
                    f"Closing writer for {name} with file {writer._file} and actual name"
                    f" {writer._file_name}"
                )
                writer.close(skip_flush=skip_flush)

    def closed_files(self, load_id: str) -> List[DataWriterMetrics]:
        """Return metrics for all fully processed (closed) files"""
        files: List[DataWriterMetrics] = []
        for name, writer in self.buffered_writers.items():
            if name.startswith(load_id):
                files.extend(writer.closed_files)

        return files

    def remove_closed_files(self, load_id: str) -> None:
        """Remove metrics for closed files in a given `load_id`"""
        for name, writer in self.buffered_writers.items():
            if name.startswith(load_id):
                writer.closed_files.clear()

    @abstractmethod
    def _get_data_item_path_template(self, load_id: str, schema_name: str, table_name: str) -> str:
        """Returns a file template for item writer. note: use %s for file id to create required template format"""
        pass



from typing import Any, Optional

from dlt.common.schema.typing import TColumnNames, TTableSchemaColumns
from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource

GEOMETRY_HINT = "x-postgres-geometry"
SRID_HINT = "x-postgres-srid"


def postgres_adapter(
    data: Any,
    geometry: TColumnNames = None,
    srid: Optional[int] = 4326,
) -> DltResource:
    """Prepares data for the postgres destination by specifying which columns should
    be cast to PostGIS geometry types.

    Args:
        data (Any): The data to be transformed. It can be raw data or an instance
            of DltResource. If raw data, the function wraps it into a DltResource
            object.
        geometry (TColumnNames, optional): Specify columns to cast to geometries.
            It can be a single column name as a string, or a list of column names.
        srid (int, optional): The Spatial Reference System Identifier (SRID) to be
            used for the geometry columns. If not provided, SRID 4326 will be used.

    Returns:
        DltResource: A resource with applied postgres-specific hints.

    Raises:
        ValueError: If input for `geometry` is invalid, or if no geometry columns are specified.

    Examples:
        >>> data = [{"town": "Null Island", "loc": "POINT(0 0)"}]
        >>> postgres_adapter(data, geometry="loc", srid=4326)
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)

    column_hints: TTableSchemaColumns = {}

    if geometry:
        if isinstance(geometry, str):
            geometry = [geometry]
        if not isinstance(geometry, list):
            raise ValueError(
                "`geometry` must be a list of column names or a single column name as a string."
            )

        for column_name in geometry:
            column_hints[column_name] = {
                "name": column_name,
                GEOMETRY_HINT: True,  # type: ignore[misc]
            }
            if srid is not None:
                column_hints[column_name][SRID_HINT] = srid  # type: ignore

    if not column_hints:
        raise ValueError("A value for `geometry` must be specified.")
    else:
        resource.apply_hints(columns=column_hints)
    return resource
